#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <assert.h>
#include <pnglite.h>

void barf(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);

	exit(1);
}

struct Image {
	png_t png;
	uint32_t *pixel_data;
};

// Access red, green, blue, and alpha component values
// in a 32-bit unsigned RGBA pixel value.
#define ALPHA(pixel)  ((pixel)>>24)
#define BLUE(pixel)   (((pixel)>>16)&0xFF)
#define GREEN(pixel)  (((pixel)>>8)&0xFF)
#define RED(pixel)    ((pixel)&0xFF)

// Encode a 32-bit unsigned RGBA value from individual
// red, green, blue, and alpha component values.
#define RGBA(r,g,b,a) ((((a) << 24)) | (((b) << 16)) | (((g) << 8)) | ((r)))

// Get the width of the image
static unsigned img_get_width( Image *img )
{
	return img->png.width;
}

static unsigned img_get_height( Image *img )
{
	return img->png.height;
}

// Get the number of pixels in an image.
static size_t img_num_pixels(Image *img)
{
	return img->png.width * img->png.height;
}

// Allocate a buffer large enough to store pixel data for given image.
uint32_t *img_alloc_buf(Image *img)
{
	size_t num_pixels = img_num_pixels(img);
	return (uint32_t *) malloc(num_pixels * sizeof(uint32_t));
}

// Get the buffer from an image that already has a buffer.
uint32_t *img_get_buf(Image *img)
{
	return img->pixel_data;
}

// Read an image from a file.
static Image *img_read(const char *filename)
{
	Image *img = (Image *) malloc(sizeof(Image));
	if (png_open_file_read(&img->png, filename) != PNG_NO_ERROR) {
		barf("Couldn't open image\n");
	}

	printf("w=%u, h=%u\n", img->png.width, img->png.height);
	printf("color type is %d\n", img->png.color_type);
	printf("bpp is %u\n", img->png.bpp);

	if (img->png.color_type != PNG_TRUECOLOR_ALPHA) {
		barf("Only true color alpha images supported\n");
	}

	img->pixel_data = img_alloc_buf(img);
	if (png_get_data(&img->png, (unsigned char *) img->pixel_data) != PNG_NO_ERROR) {
		barf("Could not read image data\n");
	}
	//printf("Read data successfully!");

	return img;
}

// Save a transformed image.
static void img_save_transformed(const char *filename, uint32_t *img_data, const Image *orig_img)
{
	png_t out;

	if (png_open_file_write(&out, filename) != PNG_NO_ERROR) {
		barf("couldn't open image to save\n");
	}

	if (png_set_data(&out,
			orig_img->png.width,
			orig_img->png.height,
			orig_img->png.depth,
			orig_img->png.color_type,
			(unsigned char *)img_data) != PNG_NO_ERROR) {
		barf("Could not save image data\n");
	}

	png_close_file(&out);
}

static void img_close_file(Image *img)
{
	png_close_file(&img->png);
}

static void img_free(Image *img)
{
	free(img->pixel_data);
	free(img);
}

__global__ void cuSwapGreenAndBlue(uint32_t *in, uint32_t *out, int w, int h)
{
	int idx = blockIdx.y * w + blockIdx.x;
	uint8_t red = RED(in[idx]);
	uint8_t green = GREEN(in[idx]);
	uint8_t blue = BLUE(in[idx]);
	uint8_t alpha = ALPHA(in[idx]);
	
	uint32_t outpixel = RGBA(red, blue, green, alpha);
	out[idx] = outpixel;
}

int main(int argc, char **argv)
{
	if (argc != 3) {
		barf("Usage: readImage <infile> <outfile>\n");
	}

	png_init(0, 0);

	Image *img = img_read(argv[1]);
	uint32_t *img_data = img_get_buf(img);

	uint32_t *out_img_data = (uint32_t *) img_alloc_buf(img);

#if 0
	size_t num_pixels = img_num_pixels(img);
	for (size_t i = 0; i < num_pixels; i++) {
		// swap blue and green values
		uint8_t red = RED(img_data[i]);
		uint8_t green = GREEN(img_data[i]);
		uint8_t blue = BLUE(img_data[i]);
		uint8_t alpha = ALPHA(img_data[i]);

		uint32_t outpixel = RGBA(red, blue, green, alpha);
		out_img_data[i] = outpixel;
	}
#else
	cudaDeviceProp prop;
	cudaGetDeviceProperties( &prop, 0 );
	printf( "%d GPU processors\n", prop.multiProcessorCount );

	size_t bufSize = img_num_pixels(img) * sizeof(uint32_t);

	uint32_t *dev_imgdata, *dev_imgdata_out;
	cudaMalloc( (void **) &dev_imgdata, bufSize );
	cudaMalloc( (void **) &dev_imgdata_out, bufSize );

	// copy image data to device
	cudaMemcpy( dev_imgdata, img_data, bufSize, cudaMemcpyHostToDevice );

	int w = img_get_width(img);
	int h = img_get_height(img);
	dim3 grid(w, h);

	cuSwapGreenAndBlue<<<grid, 1>>>( dev_imgdata, dev_imgdata_out, w, h );

	// copy transformed image data from device
	cudaMemcpy( out_img_data, dev_imgdata_out, bufSize, cudaMemcpyDeviceToHost );
#endif

	img_save_transformed(argv[2], out_img_data, img);

	img_close_file(img);
	img_free(img);
	free(out_img_data);

	return 0;
}

// vim:syntax=c
